#include <stdlib.h>
#include "gillespie.h"
#include "diffusion.h"
#include "types.h"
#include "particle_initialization.h"
#include "rng.h"
#include <assert.h>
#include <stdio.h>

extern long seed;

void populate_lattice(TLattice *lattice, TGillespie *gillespie, TGeometry *geometry, TInitConditions *init, TDiffusion *diff){
  static TLatticeLin latticeLin;
  int iSpecies;
  double siteLength;

  siteLength = 1.0 / diff->lambda;

  fprintf(stderr, "Initializing particles.");
  scan_sites(lattice, geometry, init, &latticeLin);
  fprintf(stderr, ".");

  for(iSpecies=0; iSpecies < gillespie->N; iSpecies++){
    populate_species(lattice, init, &latticeLin, gillespie->boundedSpecies, iSpecies, siteLength);
    fprintf(stderr, ".");
  }
  fprintf(stderr, " Done!\n");
}


void populate_species(TLattice *lattice,  TInitConditions *init, TLatticeLin *latticeLin, int *boundSpecies, int iSpecies, double siteLength) {
  int typeParticlesInitialization;

  typeParticlesInitialization = init->initCond[iSpecies].type;

  if( boundSpecies[iSpecies] != BOUNDED){
    init_particles_cytosol(lattice, init, latticeLin, iSpecies, typeParticlesInitialization, siteLength);
  } else {
    init_particles_membrane(lattice, init, latticeLin, iSpecies, typeParticlesInitialization, siteLength);
  }
}

void init_particles_membrane(TLattice *lattice, TInitConditions *init, TLatticeLin *latticeLin, int iSpecies, int particlesInitialization, double siteLength){
/* populates the memebrane sites with particles*/
/* membrane initial concentration is for volume concentrations of membrane species*/
  long int numberParticles;
  double volumeConcentration;

  if(particlesInitialization == TRUE){
    numberParticles = (long int) init->initCond[iSpecies].value;
  } else {
    volumeConcentration = init->initCond[iSpecies].value;
    numberParticles = rd_round ( 1000 * volumeConcentration * AVOGADRO * DX * siteLength * siteLength / 1000000000000.0); 
  }

  throw_particles(lattice, latticeLin->membrane, latticeLin->iMembraneSize, iSpecies, numberParticles);
}

void init_particles_cytosol(TLattice *lattice, TInitConditions *init, TLatticeLin *latticeLin, int iSpecies, int particlesInitialization, double siteLength){
  /*large concentration*/
  long int numberParticles;
  double siteVolume;
  double cellVolume;
  double concentration; /* in micromoles */
  int cytosolNumbSites;

 if( particlesInitialization == TRUE ){
   numberParticles = (long int) init->initCond[iSpecies].value;
 } else {
  cytosolNumbSites = latticeLin->iCytosolSize;
  concentration = init->initCond[iSpecies].value;

  siteVolume = siteLength * siteLength * siteLength; /*in micrometers*/
  cellVolume = siteVolume * cytosolNumbSites;
  numberParticles = rd_round( concentration * 6.0225*0.01 / cellVolume);
 }
  
  throw_particles(lattice, latticeLin->cytosol, latticeLin->iCytosolSize, iSpecies, numberParticles);
}

void throw_particles(TLattice *lattice, Dimension *sites, int sizeSites, int iSpecies, long int numberParticles){   
  int i,j,x,y,z;
 
  for(i=0; i < numberParticles; i++){
	/*select a cytosol site entry from the list */
    j = - 1 + ceil (rand3(&seed) * ( sizeSites - 1 + 1));
/*    j = (int) rand()/(((double)RAND_MAX + 1) / sizeSites)*/
/*    j=(int) ( (double) cytosolNumbSites * rand()/(RAND_MAX+1.0));*/
	/* obtain its coordinates */
    assert( j < sizeSites );

    x = sites[j].x;
    y = sites[j].y;
    z = sites[j].z;
	/* add one molecule */
    lattice->A[x][y][z].count[iSpecies]++;
  }
}

void scan_sites(TLattice *lattice, TGeometry *geometry, TInitConditions *init, TLatticeLin *latticeLin){
  /* Builds a list of cytosolic and membrane sites coordinates.
   * Membrane sites are also cytosolic sites.
   */
  int x,y,z;
  int ic = 0;
  int im = 0;

  assert( lattice->xdim * lattice->ydim *lattice->zdim <= MAXCYTOSOL);

  for(x=0; x< lattice->xdim; x++){
    for(y=0; y < lattice->ydim; y++){
      for( z=0; z < lattice->zdim; z++){
	switch( lattice->id[x][y][z]){
	case MEMBRANE:
	  latticeLin->membrane[im].x = x;
	  latticeLin->membrane[im].y = y;
	  latticeLin->membrane[im].z = z;
	  im++;
	case CYTOSOL:
	  latticeLin->cytosol[ic].x = x;
	  latticeLin->cytosol[ic].y = y;
	  latticeLin->cytosol[ic].z = z;
	  ic++;
	  break;
	}
      }

      /*      assert(im < MAXMEMBRANE && ic < MAXCYTOSOL);*/
    }
  }

  latticeLin->iCytosolSize = ic;
  latticeLin->iMembraneSize = im;
}
