#include "rd_simulation.h"
#include "gasdev.h"
#include "rng.h"
#include "statistics.h"
#include <assert.h>
#include "geometry.h"
#include "diffusion.h"
#include "particle_initialization.h"
#include <float.h>

extern long seed;


void reaction(TGillespie *reactionData,\
	      TLattice *lattice,\
	      double t_slot, int x, int y, int z, double lambda){
  assert(lattice->s == 0 || lattice->s == 1);

  gillespie(reactionData, &lattice->FROM[x][y][z], t_slot, lambda);

}



void set_cell_geometry(TLattice *lattice,\
			     TInitConditions *init,\
			     TGillespie *gillespie,\
			     TGeometry *geometry){
  int x, y, z;

  switch(lattice->boundary){
  case SS: /* Source sink */
  case R1: /*All species homogenously distributed */
    for(z=0; z < lattice->zdim; z++){
      for(x=0; x< lattice->xdim; x++){
	for(y=0; y<lattice->ydim; y++){
	  lattice->id[x][y][z] = CYTOSOL;
	}
      }
    }
    break;
  case R2: /*cytosol + membrane at y==0*/
    for(z=0; z < lattice->zdim; z++){   
      for(x=0; x< lattice->xdim; x++){
	for(y=0; y<lattice->ydim; y++){
	  if( y == 0 ){
	    lattice->id[x][y][z] = MEMBRANE;
	  } else {
	    lattice->id[x][y][z] = CYTOSOL;
	  }
	}
      }
    }
    /*build neighbours list for membrane*/
    /*membrane lies on plane y=0*/
    for(z=0; z < lattice->zdim; z++){
      for(x=0; x < lattice->xdim; x++){
	/*top*/
	lattice->N[x][0][z][TOP].x = (x==0)? 0: x-1;
	lattice->N[x][0][z][TOP].y = 0;
	lattice->N[x][0][z][TOP].z = z;
	/*right*/
	lattice->N[x][0][z][RIGHT].x = x;
	lattice->N[x][0][z][RIGHT].y = 0;
	lattice->N[x][0][z][RIGHT].z = (z==lattice->zdim-1)? z: z+1;
	/*bottom*/
	lattice->N[x][0][z][BOTTOM].x = (x==lattice->xdim-1)? x: x+1;
	lattice->N[x][0][z][BOTTOM].y = 0;
	lattice->N[x][0][z][BOTTOM].z = z;
	/*left*/
	lattice->N[x][0][z][LEFT].x = x;
	lattice->N[x][0][z][LEFT].y = 0;
	lattice->N[x][0][z][LEFT].z = (z==0)? 0: z-1;
      }
    }
    break;
  case GFILE: /*geometry defined in file*/
  case QUARTER:

    load_geometry(lattice, geometry);
    break;
  }

}

void initialize_lattice(TLattice *lattice, TInitConditions *init, TGillespie *gillespie, TGeometry *geometry, TDiffusion *diff){
  int x, y, z, iSpecies;

  switch(init->type){
  case SOURCESINK:
    for(z=0; z < lattice->zdim; z++){
      for(x=0; x < lattice->xdim ; x++){
	for(iSpecies=0; iSpecies<gillespie->N; iSpecies++){
	  lattice->B[x][0][z].count[iSpecies] = lattice->A[x][0][z].count[iSpecies] = init->initCond[iSpecies].value;
	  lattice->B[x][lattice->ydim - 1][z].count[iSpecies] = lattice->A[x][lattice->ydim - 1][z].count[iSpecies] = 0;
	}
      }
    }
    break;
  case HOMOGENEOUS: /* All species homogeneously (no noise) distributed */
    for(x=0; x<lattice->xdim; x++){
      for(y=0; y<lattice->ydim; y++){
	for(z=0; z<lattice->zdim; z++){
	  for(iSpecies=0; iSpecies < gillespie->N; iSpecies++){
	    if( lattice->id[x][y][z] == CYTOSOL ){
	      if( gillespie->boundedSpecies[iSpecies] != BOUNDED){
		lattice->A[x][y][z].count[iSpecies] = init->initCond[iSpecies].value;
		lattice->B[x][y][z].count[iSpecies] = 0;
	      } else {
		lattice->A[x][y][z].count[iSpecies] = 0;
		lattice->B[x][y][z].count[iSpecies] = 0;
	      }
	    } else if (lattice->id[x][y][z] == MEMBRANE ){
	      lattice->A[x][y][z].count[iSpecies] = init->initCond[iSpecies].value;
	      lattice->B[x][y][z].count[iSpecies] = 0;
	    }
	  }
	}
      }
    }
    break;
  case RANDOM:
    /* input file tells total number of particle in the cytosol or membrane */
    populate_lattice(lattice, gillespie, geometry, init, diff);
  }

  lattice->s = 0;
  lattice->FROM = lattice->A;
  lattice->TO = lattice->B;
}

void initialize_t_alarms(TGillespie *gillespie, TDiffusion *diff, double t_alarm[][2]){
  int s;
  double *D;
  double *p;

  D = diff->D;
  p = diff->p;

  for(s=0; s<gillespie->N; s++){
    if( gillespie->boundedSpecies[s] == BOUNDED ){
      t_alarm[s][0] = t_alarm[s][1] = (1.0/(2.0*(DIMENSIONALITY-1))) * (gillespie->length*gillespie->length / (diff->lambda * diff->lambda)) * (1.0 - p[0]) / (D[s]);
#ifdef DEBUG
      fprintf(stderr, "D[%d] = %e\n", s, t_alarm[s][0]);
#endif
    } else {
      t_alarm[s][0] = t_alarm[s][1] =  (1.0/(2.0*DIMENSIONALITY)) * (gillespie->length*gillespie->length / (diff->lambda * diff->lambda)) * (1.0 - p[0]) / (D[s]); /*0.25 in real 2D, 0.5 in 1D*/
#ifdef DEBUG
      fprintf(stderr, "D[%d] = %e\n", s, t_alarm[s][0]);
#endif
    }
  }
}

int rd_round(double d){
  double integer;
  double frac;

  frac = modf(d, &integer);

  if( frac >= 0.5)
    return integer + 1;
  else
    return integer;
}


void swap_lattices(TLattice *lattice){
  assert(lattice->s ==0 || lattice->s == 1);

  if( lattice->s==0){ /*current A, next B*/
    lattice->s = 1; /*current B, next A*/
    memset(lattice->A, 0, L_MAX_X_SIZE * L_MAX_Y_SIZE * L_MAX_Z_SIZE * sizeof(tSite));
    lattice->TO = lattice->A;
    lattice->FROM = lattice->B;
  } else { /*current B, next A*/
    lattice->s = 0;/*current A, next B*/
    memset(lattice->B, 0, L_MAX_X_SIZE * L_MAX_Y_SIZE * L_MAX_Z_SIZE *sizeof(tSite));
    lattice->TO = lattice->B;
    lattice->FROM = lattice->A;
  }
}

double getNext_t_sim(TGillespie *gillespie, TTimes *times, double cur_time){
  int s;
  double t_min;
  int N;
  double (*t_alarm)[2];
  double *t_tau;

  t_alarm = times->t_alarm;
  t_tau = &times->t_tau;

  N = gillespie->N;
  t_min = t_alarm[0][0];
  /*  *t_tau = t_alarm[0][1];*/
  *t_tau = t_alarm[0][0] - cur_time;

  if( N >1 ){
    for(s=1; s<N; s++){
      if(t_min < t_alarm[s][0]){
	t_min = t_min;
      } else {
	t_min = t_alarm[s][0];
	*t_tau = t_alarm[s][0] - cur_time;
      }
    }
  }  

  return t_min;
}

void print_parameters(TLattice *lattice, TDiffusion *diff,\
		      TGillespie *gillespie, TTimes *times,\
		      TInitConditions *init){
  int x;

  fprintf(stderr, "N = %d, M = %d\n", gillespie->N, gillespie->M);
  for(x=0;x<gillespie->N; x++){
    fprintf(stderr, "D[%d] = %lf\n", x, diff->D[x]);
  }
  for(x=0; x< gillespie->M; x++){
    fprintf(stderr, "c[%d] = %lf\n", x, gillespie->c[x]);
  }
  for(x=0; x < gillespie->N; x++){
    fprintf(stderr, "Init[%d] = %lf\n",x,init->initCond[x].value);
  }
  
  fprintf(stderr, "Lattice size L= %d, %d, %d.\n", lattice->xdim, lattice->ydim, lattice->zdim);
}

/* special function for simulating ex9.3*/
int particles_to_print(TLattice *lattice, double t){
  static int limit = -1;
  static int count = 0;
  long unsigned int *acc;
  /*we only need to reach type B particles, 2nd index*/
  acc = getMass(lattice, 2);

  if( limit == -1){
    limit = acc[1];
    print_system_mass(lattice, 3, t);
    count++;
  } else {
    if( acc[1] < ceil(14* limit / 16) ){
      print_system_mass(lattice, 3, t);
      limit = ceil(14*limit/ 16); 
      count++;
    }
  }

  return (count>2)?TRUE:FALSE; /*after max 3 measurements quit.*/
}


void rd_simulation(TLattice *lattice,\
		   TInitConditions *init,\
		   TGillespie *gillespie,\
		   TOutput *output,
		   TDiffusion *diff,
		   TTimes *times,
		   TGeometry *geometry,
		   int experiment)
{
  int x,y,z;
  double t; /*current simulation time*/
  int timeIndex;
  int iSpecies;
  double pt;

  double t_max; /*maximum simulation time*/
  double tau_print; /*print interval time*/
  double t_print; /*next print alarm*/
  double (*t_alarm)[2]; /*Diffusion alarm times*/

  /*  geometry->filename = NULL;*/

  print_parameters(lattice, diff, gillespie, times, init);

  t_max = times->t_max;
  tau_print = times->tau_print;
  t_print = times->t_print;
  t_alarm = times->t_alarm;
  t = 0.0;

  initialize_t_alarms(gillespie, diff, t_alarm);

  memset(lattice->A, 0, L_MAX_X_SIZE*L_MAX_Y_SIZE*L_MAX_Z_SIZE*sizeof(tSite));
  memset(lattice->B, 0, L_MAX_X_SIZE*L_MAX_Y_SIZE*L_MAX_Z_SIZE *sizeof(tSite));
  set_cell_geometry(lattice, init, gillespie, geometry);
  initialize_lattice(lattice, init, gillespie, geometry, diff);

  if(gillespie->output->vis_method == 4){
    t_print=DBL_MAX;
  } else {
    t_print= tau_print;
  }

  pt = 0.0;


  t = getNext_t_sim(gillespie, times, t);
  timeIndex = 0;

  for(; t<t_max; t = getNext_t_sim( gillespie, times, t), timeIndex++){
#ifdef DEBUG
    fprintf(stderr, "Iter. %8lf\n", t);
#endif
#ifdef COLLISION
    fprintf(stderr, "0\n");
#endif

    gillespie->simTime = t;

    for(z=0; z < lattice->zdim; z++){
      for(y=0; y < lattice->ydim; y++){
	for(x=0; x < lattice->xdim; x++){
	  /*reaction*/
	  reaction(gillespie, lattice, times->t_tau, x, y, z, diff->lambda);
	  
	  /*diffusion*/
	  diffusion(diff, x, y, z, lattice, gillespie, t_alarm, t);
	  
	}
      }
    }

    /*advance t_alarm for diffused species*/
    for(iSpecies=0; iSpecies < gillespie->N; iSpecies++){
      if(t_alarm[iSpecies][0] == t){
	t_alarm[iSpecies][0] += t_alarm[iSpecies][1];
      }
    }

    swap_lattices(lattice);

    if( t >= t_print){
      switch(output->print_method){
      case 0: print_profile(lattice,lattice->A, lattice->B,t, lattice->boundary, gillespie->N); break;
      case 1: print_particles(lattice,gillespie->N, t); break;
      case 2: print_system_mass(lattice,gillespie->N, t);break;
      case 3: print_2d(lattice, gillespie);break;
      case 4: print_radius_profile(lattice, gillespie, t, geometry); break;
      case 5: print_matrix(lattice, gillespie, t);
      default:; 
      }

      switch(output->stat_method){
      case 0: break;
      case 1: break;
      case 2: 
	add_st_mass(output->statistics.statMass, experiment, timeIndex, getMass(lattice,gillespie->N), t, gillespie->N); break;
      default:;
      }

      switch(output->vis_method){
      case 0:
	t_print += tau_print;
	break;
      case 1:
	t_print = tau_print * pow(10.0,pt);
	pt+= 1.0;
	break;
      case 2: /*print initial and last*/
	t_print = times->t_max;
	break;
      case 3: /*print at certain number of particles and quit*/
	if(particles_to_print(lattice, t)){
	  /*quit simulation*/
	  exit(1);
	}
	break;
      case 4:
	/*printing code in Gillespie method*/
      default:;
      }
    }

  } /* end simulation */

  if( t >= t_print){
    switch(output->print_method){
    case 0: print_profile(lattice,lattice->A, lattice->B, t, lattice->boundary, gillespie->N); break;
    case 1: print_particles(lattice,gillespie->N, t); break;
    case 2: print_system_mass(lattice,gillespie->N,t);break;
    case 3: print_2d(lattice, gillespie);break;
    case 4: print_radius_profile(lattice, gillespie, t, geometry); break;
    case 5: print_matrix(lattice, gillespie, t);
    default:; 
    }
  }
}


