#include "read_configuration.h"
#include "types.h"
#include <string.h>
#include "gasdev.h"
#include <sys/types.h>
#include <unistd.h>

extern long seed;

void read_configuration(char *filename,\
			TGillespie *gillespie,\
			TLattice *lattice,\
			TInitConditions *init,\
			TOutput *output,\
			TDiffusion *diffusion,\
			TTimes *times,\
			TGeometry *geometry){
  FILE *f;
  int x;
  char fln[20];
  double p_i;
  int version;

  tSite (*A)[L_MAX_Y_SIZE][L_MAX_Z_SIZE];
  double *p;
  double tau; /*not used. just to keep compatibility*/
  double *t_max;
  double *t_print;
  TInitPart *initialValue;
  eBoundary *boundary;
  int *N;
  double *D;
  int *M;
  /* version 6 */
  int bs; /* number of membrane-bounded speceis*/
  int bsi; /*0-based index of membrane-boudned species*/
  
  A = lattice->A;
  p = diffusion->p;
  D = diffusion->D;
  initialValue = init->initCond;
  boundary = &lattice->boundary;
  N = &gillespie->N;
  M = &gillespie->M;
  t_max = &times->t_max;
  t_print = &times->t_print;


  f= fopen(filename, "r");
  if( f == NULL){
    fprintf(stderr, "Cannot open input file %s\n", filename);
    exit(-1);
  }

  fscanf(f, "#%d\n", &version);
  if( version != FILE_VERSION){
    fprintf(stderr, "\n\nInput file version %d is not correct.\n", version);
    exit(-1);
  }

  fscanf(f, "%d, %d\n", N, M);
  if( *N > MAXSPECIES){
    fprintf(stderr, "Max number of species overpassed.\n");
    exit(-1);
  }

  /*read p_0, the rest of p_i are (1-p_0)/4*/
  fscanf(f, "%s\n", fln);
  p[0] = atof(fln);
  
  p_i = (1.0 - p[0]) / (2*DIMENSIONALITY);

#ifdef DEBUG
  fprintf(stderr, "p_0 = %f,\t p_i = %f\n", p[0], p_i);
#endif

  for(x=1; x<NEIGH; x++){
    p[x] = p[x-1] + p_i;
  }

  if( p[x-1] < 1.0){
    fprintf(stderr, "%d-th prob. is %lf. Addjusting to be 1.0\n", x-1,p[x-1]);
    p[x-1] = 1.0;
  }


  /* modification for version 9
  read type of initial value
     0 - concentration
     1 - particles
     then value
  */

  for(x=0; x<*N; x++)
    fscanf(f, "%d %lf\n", &initialValue[x].type, &initialValue[x].value);


  /*type of boundary condition*/
  fscanf(f, "%d\n", (int*)boundary);
#ifdef DEBUG
  fprintf(stderr, "using boundary condition: %d\n", *boundary);
#endif


  fscanf(f, "%s\n", fln);
  *t_max = atof(fln);

#ifdef DEBUG
  fprintf(stderr, "Simulation time: %lf (units of time)\n", *t_max);
#endif

  fscanf(f, "%s\n", fln);
  tau = atof(fln);

  fscanf(f, "%s\n", fln);
  *t_print = atof(fln);
#ifdef DEBUG
  fprintf(stderr, "Output interval: %lf (units of time)\n", *t_print);
#endif

  fscanf(f, "%li\n", &seed);
  if( seed == 0){
    seed = -1 * getpid();
    srand48((-1)*seed);
    initGaussTable();
    fprintf(stderr, "Undefined RNG seed= %li\n", seed);
  }else{
    fprintf(stderr, "Defined RNG seed= %li\n", seed);
  }

  /*Diffusion coefficients*/
  for(x=0; x< *N; x++){
    fscanf(f, "%lf\n",&D[x]);
  }

  /* Reaction Data */
  gillespie->c = malloc(sizeof(double)* (*M));
  for(x=0; x< *M; x++){
    fscanf(f, "%lf\n", &gillespie->c[x]);
#ifdef DEBUG
    fprintf(stderr, "c[%d]= %g\n", x, gillespie->c[x]);
#endif
  }
  
  gillespie->cX = malloc(sizeof(int) * (*M) * 10);
  for(x=0; x< *M; x++){
    fscanf(f, "%d,%d,%d,%d,%d,%d,%d,%d,%d,%d\n", &gillespie->cX[x][0],\
	   &gillespie->cX[x][1], &gillespie->cX[x][2],\
	   &gillespie->cX[x][3], &gillespie->cX[x][4],\
	   &gillespie->cX[x][5], &gillespie->cX[x][6],\
	   &gillespie->cX[x][7], &gillespie->cX[x][8], &gillespie->cX[x][9]);
#ifdef DEBUG
    fprintf(stderr, "cX = %d,%d,%d,%d,%d,%d,%d,%d,%d,%d\n",gillespie->cX[x][0],\
	    gillespie->cX[x][1], gillespie->cX[x][2],\
	    gillespie->cX[x][3], gillespie->cX[x][4],\
	    gillespie->cX[x][5], gillespie->cX[x][6],\
	    gillespie->cX[x][7], gillespie->cX[x][8], gillespie->cX[x][9]); 
#endif
  }

  gillespie->N = *N;
  gillespie->M = *M;

  /*version 6*/
  fscanf(f, "%d\n", &bs);
  
  for(x=0; x<MAXSPECIES; x++){
    gillespie->boundedSpecies[x] = 0;
  }
  for(x=0; x<bs; x++){
    fscanf(f, "%d\n", &bsi);
    gillespie->boundedSpecies[bsi] = BOUNDED;
  }

  /* READ LATTICE SIZE INFORMATION */

  fscanf(f,"%d,%d,%d\n", &(lattice->xdim), &(lattice->ydim), &(lattice->zdim));

  if( lattice->xdim > L_MAX_X_SIZE || lattice->ydim > L_MAX_Y_SIZE || lattice->zdim > L_MAX_Z_SIZE){
    fprintf(stderr, "Lattice size larger than maximum accepted %d x %d x %d. \n",\
	    L_MAX_X_SIZE,L_MAX_Y_SIZE, L_MAX_Z_SIZE);
    exit(-1);
  }

  /* read lambda parameter for diffusion */

  fscanf(f,"%lf\n", &diffusion->lambda);

  fscanf(f, "%s\n", geometry->filename);
  fprintf(stderr, "Using %s geometry.\n", geometry->filename);

  /*********************** version 9 extensions *********************************/

  /*Initialization method:
    0 - Homogeneous: concentration or particle is the same in each lattice site
    1 - Stochastic (or Random) : throw particles randomly
  */

  fscanf(f, "%s\n", fln);
  if( strcmp(fln, "stochastic") == 0 ){
    init->type = 1;
  } else {
    if (strcmp(fln, "homogeneous")){
      init->type = 0;
    } else{
      fprintf(stderr, "Initialization method option \"%s\" not valid. 0 - Homogeneous, 1 - Stochastic\n", fln); 
      exit(-1);
    }
  }

  /*verison 12*/
  fscanf(f, "%lf\n", &gillespie->length);

  fclose(f);
  
}
