#include"statistics.h"

void statistics(int j, double value, double *cur_mean, double *cur_var){
  double next_mean, next_var;
  double mean, var;

  mean = *cur_mean;
  var = *cur_var;

  next_mean = mean + (value - mean) / (j + 1.0);
  if(j>0)
    next_var = (1.0 - (1.0/j)) * var + (j+1.0) * (value - mean) * (value - mean);
  else
    next_var = 0.0;
  
  *cur_mean = next_mean;
  *cur_var = next_var;
}

long unsigned int *getMass(TLattice *lattice, int N){
  int x,y,z,sp;
  static long unsigned int acc[MAXSPECIES];

  for(sp=0; sp<N; sp++){
    acc[sp] = 0;
    for( z =0; z < lattice->zdim; z++){
      for(y=0; y<lattice->ydim; y++){
	for(x=0; x<lattice->xdim; x++){
	  acc[sp] += lattice->FROM[x][y][z].count[sp];
	}
      }
    }
  }
  return acc;
}

void print_system_mass(TLattice *lattice, int N, double t){
  int sp;
  long unsigned int *acc;

  printf("%.5e\t", t);

  acc = getMass(lattice, N);

  for(sp=0; sp<N; sp++){
      printf("%lu\t", acc[sp]);
  }  
  printf("\n");
#ifdef DEBUG
  fflush(stdout);
#endif

}

void print_particles(TLattice *lattice, int N, double t){
  int x,y,z,sp;

  printf("#time %.4e\n", t);

  for(z=0; z<lattice->zdim; z++){
    for(y=0; y<lattice->ydim; y++){
      for(x=0; x<lattice->xdim; x++){
	if(lattice->id[x][y][z] != EXTERIOR){
	  printf("%d\t%d\t%d\t", x,y,z);
	  for(sp=0; sp<N; sp++){
	    printf("%d\t", lattice->FROM[x][y][z].count[sp]);
	  }
	  putchar('\n');
	}
      }
    }
    putchar('\n');
  }
}

void print_2d(TLattice *lattice, TGillespie *gillespie){
  int x,y,z,i;

  z = 0;
    for(x=0; x < lattice->xdim; x++){
      for(y=0; y < lattice->ydim; y++){
	printf("%d %d ", x, y);
	for(i=0; i < gillespie->N; i++){
	  if(lattice->s==0){
	    printf("%d ", lattice->A[x][y][z].count[i]);
	  }else {
	    printf("%d ", lattice->B[x][y][z].count[i]);
	  }	   
	}
	printf("\n");
      }
      printf("\n");
    }
    
    printf("\n\n");

}


void print_profile(TLattice *lattice, tSite A[][L_MAX_Y_SIZE][L_MAX_Z_SIZE], tSite B[][L_MAX_Y_SIZE][L_MAX_Z_SIZE], double t, eBoundary boundary, int N){
  int v, y, x,z, sp;
  int s;
  double mean[MAXSPECIES], var[MAXSPECIES];

  printf("#time (units) %f\n", t);
  s = lattice->s;

  for(y=0; y<lattice->ydim; y++){
    for(sp=0; sp<N; sp++){
      v = 0;
      memset( mean, 0, sizeof(double) * MAXSPECIES);
      memset(var, 0, sizeof(double) * MAXSPECIES);

      for(z=0; z<lattice->zdim; z++){
	for(x=0; x<lattice->xdim; x++){
	  v = (s==1)? B[x][y][z].count[sp]: A[x][y][z].count[sp];
	  statistics((x*z)+z, v, &mean[sp], &var[sp]);
	}
      }
      /*print row, mean, standard error of the mean*/
      printf("%d\t%lf\t%lf\t", y, mean[sp], sqrt(var[sp]/(lattice->xdim*lattice->zdim)));
      /*    for(x=0; x<L_MAX_X_SIZE; x++)
	    printf("%d  ", (s==1)? B[x][y]: A[x][y]);
	    printf("\n");
      */
    }
    printf("\n");
  }
  
  printf("\n\n\n");
#ifdef DEBUG
  fflush(stdout);
#endif
}

void **malloc_2d(int d1, int d2, int element_size){
  int i;
  void **u;

  u = malloc(sizeof(void *) * d1 );
  if( u == NULL){
    fprintf(stderr, "Cannot allocate memory.\n");
    exit(-1);
  }

  for(i=0; i<d1; i++){
    u[i] = malloc(element_size * d2);
    if( u[i] == NULL){
      fprintf(stderr, "Cannot allocate memory.\n");
      exit(-1);
    }
  }

  return u;
}

void ***malloc_3d(int d1, int d2, int d3, int element_size){
  int x,y;
  void ***u;

  u = (void ***)malloc_2d(d1, d2, sizeof(void *));
  for(x=0;x<d1;x++){
    for(y=0;y<d2;y++){
      u[x][y] = malloc(element_size * d3);
      if( u[x][y] == NULL){
	fprintf(stderr, "Cannot allocate memory.\n");
	exit(-1);
      }
    }
  }
  return u;
}

void free_2d(void **u, int d1){
  int i;

  for(i=0; i<d1; i++){
    free(u[i]);
  }
  free(u);
}

void free_3d(void ***u, int d1, int d2){
  int i;

  for(i=0; i<d1; i++){
    free_2d(u[i], d2);
  }
  free(u);
}

/*
 * Multiexperiment statistics. Average, Std deviation for system's mass.
 */
double ***initialize_st_mass(double simTime, double timeStep, int nSpecies){
  int ns; /*number of samples in time */
  double ***u;
  int x,y,z;

  ns = (int)ceil(simTime / timeStep) + 1;
  u = (double ***)malloc_3d(ns, 3, nSpecies, sizeof(double));
  for(x=0;x<ns;x++)
    for(y=0;y<3;y++)
      for(z=0;z<nSpecies;z++)
	u[x][y][z] = 0;

  return u;
}

void add_st_mass(double ***data, int iteration, int timeIndex, long unsigned int *mass, double time, int nSpecies){
  int i;
  for(i=0; i<nSpecies; i++){
    statistics(iteration, mass[i], &data[timeIndex][1][i], &data[timeIndex][2][i]);
    data[timeIndex][0][i] = time; /*inneficient, yet simple*/
    fprintf(stderr, "time %lf \n", time);
  }
}

void write_st_mass(double ***data, double simTime, double timeStep, int nSpecies){
  int ns;
  int t,s;
  double time, mean, stdem;
  int initial_zero_time = 0;

  ns = (int)ceil(simTime / timeStep) + 1;
  for(t=0; t<ns; t++){
    time = data[t][0][0];

    if( (time == 0.0 || t ==0 )&& initial_zero_time == 0){
      initial_zero_time = 1;
    } else if(time == 0 && t != 0){
      break;
    }

    printf("%lf\t", time);

    for(s=0; s<nSpecies; s++){
      mean = data[t][1][s];
      stdem = data[t][2][s];
      printf("%lf\t%lf\t",mean, stdem);
    }
    printf("\n");
  }

  /** free array memory */
}


void print_radius_profile(TLattice *lattice, TGillespie * gillespie, double t, TGeometry *geom){
  int x,y,i,z, xx, yy, zz;
  float r;
  int cx,cy,cz;

  cx = geom->centre.x;
  cy = geom->centre.y;
  cz = geom->centre.z;

  printf("#time %.4e\n", t);
  for(z=0; z<lattice->zdim; z++){
    zz = (z-cz)*(z-cz);
    for(y=0; y<lattice->ydim; y++){
      yy = (y-cy)*(y-cy);
      for(x=0; x< lattice->xdim; x++){
	if( lattice->id[x][y][z] != EXTERIOR){
	  xx = (x-cx)*(x-cx);
	  r = sqrt(xx + yy + zz);
	  printf("%.4f ", r);
	  for(i=0; i < gillespie->N; i++){
	    printf("%d ", lattice->FROM[x][y][z].count[i]);
	  }
	  printf("\n");
	}
      }
    }
  }

  printf("\n\n\n"); 
  }


void print_matrix(TLattice *la, TGillespie *gi, double t){
  int x,y,z,i;

  z = 0;
  /* file format to be read from Mathematica's Import Table function */
  /*  for(z=0; z<la->zdim; z++){*/
    for(i=0; i < gi->N; i++){
      /*    printf("{");*/
      for(x=0; x < la->xdim; x++){
	/*      if(x>0)printf(",");
		printf("{");*/
	for(y=0; y < la->ydim; y++){
	  /*	if(y>0) printf(", ");*/
	  printf("%d ", la->FROM[x][y][z].count[i]);
	}	
	printf("\n");
	
      }
      printf("\n");
    }
    printf("\n\n");
    /*  }
  
  printf("\n\n\n");*/
}



